# Needed for NotRequired with Python 3.7 - 3.9
# See https://www.python.org/dev/peps/pep-0655/#usage-in-python-3-11
from __future__ import annotations

import ast
import copy
import functools
import json
import os
import sys
from pathlib import Path
from textwrap import dedent
from typing import Callable, Dict, Iterable, List, Literal, Set, Union

# Even though TypedDict is available in Python 3.8, because it's used with NotRequired,
# they should both come from the same typing module.
# https://peps.python.org/pep-0655/#usage-in-python-3-11
if sys.version_info >= (3, 11):
    from typing import NotRequired, TypedDict
else:
    from typing_extensions import NotRequired, TypedDict

from ._assets import shinylive_assets_dir, repodata_json_file, ensure_shinylive_assets
from ._app_json import FileContentJson

# Files in Pyodide that should always be included.
BASE_PYODIDE_FILES = {
    "pyodide_py.tar",
    "pyodide.asm.js",
    "pyodide.asm.data",
    "pyodide.asm.wasm",
    "repodata.json",
}

# Packages that should always be included in a Shinylive deployment.
BASE_PYODIDE_PACKAGES = {"distutils", "micropip"}


# =============================================================================
# Data structures used in pyodide/repodata.json
# =============================================================================
# Note: This block of code is copied from /scripts/pyodide_packages.py
class PyodidePackageInfo(TypedDict):
    name: str
    version: str
    file_name: str
    install_dir: Literal["lib", "site"]
    sha256: str
    depends: List[str]
    imports: List[str]
    unvendored_tests: NotRequired[bool]


# The package information structure used by Pyodide's repodata.json.
class PyodideRepodataFile(TypedDict):
    info: Dict[str, str]
    packages: Dict[str, PyodidePackageInfo]


# =============================================================================
# HTML Dependency types
# =============================================================================
class HtmlDepItem(TypedDict):
    name: str
    path: str
    attribs: NotRequired[Dict[str, str]]


class QuartoHtmlDependency(TypedDict):
    name: str
    version: NotRequired[str]
    scripts: NotRequired[List[Union[str, HtmlDepItem]]]
    stylesheets: NotRequired[List[Union[str, HtmlDepItem]]]
    resources: NotRequired[List[HtmlDepItem]]


# =============================================================================
# Conversion functions
# =============================================================================
def _dep_names_to_pyodide_pkg_infos(
    dep_names: Iterable[str],
) -> List[PyodidePackageInfo]:
    repodata = _pyodide_repodata()
    pkg_infos: List[PyodidePackageInfo] = [
        copy.deepcopy(repodata["packages"][dep_name]) for dep_name in dep_names
    ]
    return pkg_infos


def _pyodide_pkg_info_to_quarto_html_dep(
    pkg: PyodidePackageInfo,
    path_prefix: str,
) -> QuartoHtmlDependency:
    """
    Convert a PyodidePackageInfo object to a QuartoHtmlDependency object.
    """
    dep: QuartoHtmlDependency = {
        "name": pkg["name"],
        "resources": [
            {
                "name": pkg["file_name"],
                "path": os.path.join(path_prefix, pkg["file_name"]),
            }
        ],
    }

    return dep


def _pyodide_pkg_infos_to_quarto_html_deps(
    pkgs: List[PyodidePackageInfo],
    path_prefix: str,
) -> List[QuartoHtmlDependency]:
    return [_pyodide_pkg_info_to_quarto_html_dep(pkg, path_prefix) for pkg in pkgs]


# =============================================================================
# Shinylive base dependencies
# =============================================================================
def shinylive_base_deps_htmldep(
    path_prefix: str = "shinylive-dist/",
) -> QuartoHtmlDependency:
    """
    Return an HTML dependency object consisting of files that are base dependencies; in
    other words, the files that are always included in a Shinylive deployment.
    """

    # First, get the list of base files.
    base_files = shinylive_base_files()

    # Next, categorize the base files into scripts, stylesheets, and resources.
    scripts: List[Union[str, HtmlDepItem]] = []
    stylesheets: List[Union[str, HtmlDepItem]] = []
    resources: List[HtmlDepItem] = []

    for file in base_files:
        if os.path.basename(file) in [
            "load-serviceworker.js",
            "jquery.min.js",
            "jquery.terminal.min.js",
            "run-python-blocks.js",
        ]:
            script_item: HtmlDepItem = {
                "name": file,
                "path": os.path.join(path_prefix, file),
            }

            if os.path.basename(file) in [
                "load-serviceworker.js",
                "run-python-blocks.js",
            ]:
                script_item["attribs"] = {"type": "module"}

            scripts.append(script_item)

        if os.path.basename(file) in [
            "jquery.terminal.min.css",
            "shinylive.css",
        ]:
            stylesheets.append(
                {
                    "name": file,
                    "path": os.path.join(path_prefix, file),
                }
            )
        else:
            resources.append(
                {
                    "name": file,
                    "path": os.path.join(path_prefix, file),
                }
            )

    # Sort scripts so that load-serviceworker.js is first, and run-python-blocks.js is
    # last.
    def scripts_sort_fun(x: Union[str, HtmlDepItem]) -> int:
        if isinstance(x, str):
            filename = x
        else:
            filename = os.path.basename(x["name"])

        if filename == "load-serviceworker.js":
            return 0
        elif filename == "run-python-blocks.js":
            return 2
        else:
            return 1

    scripts.sort(key=scripts_sort_fun)

    return {
        "name": "shinylive-base",
        "scripts": scripts,
        "stylesheets": stylesheets,
        "resources": resources,
    }


def shinylive_base_files() -> List[str]:
    """
    Return a list of files that are base dependencies; in other words, the files that are
    always included in a Shinylive deployment.
    """
    ensure_shinylive_assets()

    base_files: List[str] = []
    for root, dirs, files in os.walk(shinylive_assets_dir()):
        root = Path(root)
        rel_root = root.relative_to(shinylive_assets_dir())
        if rel_root == Path("."):
            dirs.remove("scripts")
            dirs.remove("export_template")
        elif rel_root == Path("shinylive"):
            files.remove("examples.json")
        elif rel_root == Path("shinylive/pyodide"):
            dirs.remove("fonts")
            files[:] = BASE_PYODIDE_FILES

        for file in files:
            if file.startswith("."):
                continue
            base_files.append(str(rel_root / file))

    return base_files


# =============================================================================
# Find which packages are used by a Shiny application
# =============================================================================
def package_deps_htmldep(
    json_file: Union[str, Path],
    path_prefix: str = "shinylive-dist/",
    verbose: bool = True,
) -> List[QuartoHtmlDependency]:
    """
    Find package dependencies from an app.json file, and return as a list of
    QuartoHtmlDependency objects.
    """

    def verbose_print(*args: object) -> None:
        if verbose:
            print(*args)

    json_file = Path(json_file)

    with open(json_file) as f:
        file_contents: List[FileContentJson] = json.load(f)

    pkg_infos = find_package_deps(file_contents)
    deps = _pyodide_pkg_infos_to_quarto_html_deps(pkg_infos, path_prefix)
    return deps


def base_package_deps() -> List[PyodidePackageInfo]:
    """
    Return list of packages that should be included in all Shinylive deployments. The
    returned data structure is a list of PyodidePackageInfo objects.
    """
    dep_names = _find_recursive_deps(BASE_PYODIDE_PACKAGES)
    pkg_infos = _dep_names_to_pyodide_pkg_infos(dep_names)

    return pkg_infos


def find_package_deps(
    app_contents: List[FileContentJson],
    verbose_print: Callable[..., None] = lambda *args: None,
) -> List[PyodidePackageInfo]:
    """
    Find package dependencies from the contents of an app.json file. The returned data
    structure is a list of PyodidePackageInfo objects.
    """

    imports: Set[str] = _find_import_app_contents(app_contents)

    # TODO: Need to also add in requirements.txt, and find dependencies of those
    # packages, in case any of those dependencies are included as part of pyodide.
    verbose_print("Imports detected in app:\n ", ", ".join(sorted(imports)))

    dep_names = _find_recursive_deps(imports, verbose_print)
    pkg_infos = _dep_names_to_pyodide_pkg_infos(dep_names)

    return pkg_infos


# =============================================================================
# Internal functions
# =============================================================================
def _find_recursive_deps(
    pkgs: Iterable[str],
    verbose_print: Callable[..., None] = lambda *args: None,
) -> List[str]:
    """
    Given a list of packages, recursively find all dependencies that are contained in
    repodata.json. This returns a list of all dependencies, including the original
    packages passed in.
    """
    repodata = _pyodide_repodata()
    deps = list(pkgs)
    i = 0
    while i < len(deps):
        dep = deps[i]
        if dep not in repodata["packages"]:
            # TODO: Need to distinguish between built-in packages and external ones in
            # requirements.txt.
            verbose_print(
                f"  {dep} not in repodata.json. Assuming it is in base Pyodide or in requirements.txt."
            )
            deps.remove(dep)
            continue

        dep_deps = set(repodata["packages"][dep]["depends"])
        new_deps = dep_deps.difference(deps)
        deps.extend(new_deps)
        i += 1

    return deps


def _dep_name_to_dep_file(dep_name: str) -> str:
    """
    Given the name of a dependency, like "pandas", return the name of the .whl file,
    like "pandas-1.4.2-cp310-cp310-emscripten_3_1_14_wasm32.whl".
    """
    repodata = _pyodide_repodata()
    return repodata["packages"][dep_name]["file_name"]


def _dep_names_to_dep_files(dep_names: List[str]) -> List[str]:
    """
    Given a list of dependency names, like ["pandas"], return a list with the names of
    corresponding .whl files (from data in repodata.json), like
    ["pandas-1.4.2-cp310-cp310-emscripten_3_1_14_wasm32.whl"].
    """
    repodata = _pyodide_repodata()
    dep_files = [repodata["packages"][x]["file_name"] for x in dep_names]
    return dep_files


def _find_import_app_contents(app_contents: List[FileContentJson]) -> Set[str]:
    """
    Given an app.json file, find packages that are imported.
    """
    imports: Set[str] = set()
    for file_content in app_contents:
        if not file_content["name"].endswith(".py"):
            continue

        imports = imports.union(_find_imports(file_content["content"]))

    return imports


@functools.lru_cache
def _pyodide_repodata() -> PyodideRepodataFile:
    """
    Read in the Pyodide repodata.json file and return the contents. The result is
    cached, so if the file changes, it won't register until the Python session is
    restarted.
    """
    with open(repodata_json_file(), "r") as f:
        return json.load(f)


# From pyodide._base.find_imports
def _find_imports(source: str) -> List[str]:
    """
    Finds the imports in a Python source code string

    Parameters
    ----------
    source : str
       The Python source code to inspect for imports.

    Returns
    -------
    ``List[str]``
        A list of module names that are imported in ``source``. If ``source`` is not
        syntactically correct Python code (after dedenting), returns an empty list.

    Examples
    --------
    >>> from pyodide import find_imports
    >>> source = "import numpy as np; import scipy.stats"
    >>> find_imports(source)
    ['numpy', 'scipy']
    """
    # handle mis-indented input from multi-line strings
    source = dedent(source)

    try:
        mod = ast.parse(source)
    except SyntaxError:
        return []
    imports: Set[str] = set()
    for node in ast.walk(mod):
        if isinstance(node, ast.Import):
            for name in node.names:
                node_name = name.name
                imports.add(node_name.split(".")[0])
        elif isinstance(node, ast.ImportFrom):
            module_name = node.module
            if module_name is None:
                continue
            imports.add(module_name.split(".")[0])
    return list(sorted(imports))



import filecmp
import os
import shutil
import sys
from pathlib import Path
from typing import Callable, List, Union


def is_relative_to(path: Path, base: Path) -> bool:
    """
    Wrapper for `PurePath.is_relative_to`, which was added in Python 3.9.
    """
    if sys.version_info >= (3, 9):
        return path.is_relative_to(base)
    else:
        try:
            path.relative_to(base)
            return True
        except ValueError:
            return False


def path_length(path: Union[str, Path]) -> int:
    """Returns the number of elements in a path.

    For example 'a' has length 1, 'a/b' has length 2, etc.
    """

    path = str(path)
    if os.path.isabs(path):
        raise ValueError("path must be a relative path")

    # Unfortunately, there's no equivalent of os.path.normpath for Path objects.
    path = os.path.normpath(path)
    if path == ".":
        return 0

    # On Windows, replace backslashes with forward slashes.
    if os.name == "nt":
        path.replace("\\", "/")

    return len(path.split("/"))


def listdir_recursive(dir: Union[str, Path]) -> List[str]:
    dir = Path(dir)
    all_files: List[str] = []

    for root, _dirs, files in os.walk(dir):
        root = Path(root)
        rel_root = root.relative_to(dir)

        for file in files:
            all_files.append(os.path.join(rel_root / file))

    return all_files


def copy_file_and_substitute(
    src: Union[str, Path], dest: Union[str, Path], search_str: str, replace_str: str
) -> None:
    with open(src, "r") as fin:
        in_content = fin.read()
        in_content = in_content.replace(search_str, replace_str)
        with open(dest, "w") as fout:
            fout.write(in_content)


def create_copy_fn(
    overwrite: bool, verbose_print: Callable[..., None] = lambda *args: None
) -> Callable[..., None]:
    """Returns a function that can be used as a copy_function for shutil.copytree.

    If overwrite is True, the copy function will overwrite files that already exist.
    If overwrite is False, the copy function will not overwrite files that already exist.
    """

    def copy_fn(src: str, dst: str, **kwargs: object) -> None:
        if os.path.exists(dst):
            if filecmp.cmp(src, dst) is False:
                print(
                    "\nSource and destination copies differ:",
                    dst,
                    """\nThis is probably because your shinylive sources have been updated and differ from the copy in the exported app.""",
                    """\nYou probably should remove the export directory and re-export the application.""",
                )
            if overwrite:
                verbose_print(f"Overwriting {dst}")
                os.remove(dst)
            else:
                verbose_print(f"Skipping {dst}")
                return

        shutil.copy2(src, dst, **kwargs)

    return copy_fn



from typing import List, Union, Optional
import os
from pathlib import Path
import re
import shutil
import urllib.request

from ._version import SHINYLIVE_ASSETS_VERSION


def download_shinylive(
    destdir: Union[str, Path, None] = None,
    version: str = SHINYLIVE_ASSETS_VERSION,
    url: Optional[str] = None,
) -> None:
    import tarfile

    if destdir is None:
        # Note that this is the cache directory, which is the parent of the assets
        # directory. The tarball will have the assets directory as the top-level subdir.
        destdir = shinylive_cache_dir()

    if url is None:
        url = shinylive_bundle_url(version)

    destdir = Path(destdir)
    tmp_name = None

    try:
        print(f"Downloading {url}...")
        tmp_name, _ = urllib.request.urlretrieve(url)

        print(f"Unzipping to {destdir}/")
        with tarfile.open(tmp_name) as tar:
            tar.extractall(destdir)
    finally:
        if tmp_name is not None:
            Path(tmp_name).unlink(missing_ok=True)


def shinylive_bundle_url(version: str = SHINYLIVE_ASSETS_VERSION) -> str:
    """
    Returns the URL for the Shinylive assets bundle.
    """
    return (
        "https://github.com/rstudio/shinylive/releases/download/"
        + f"v{SHINYLIVE_ASSETS_VERSION}/shinylive-{SHINYLIVE_ASSETS_VERSION}.tar.gz"
    )


def shinylive_cache_dir() -> str:
    """
    Returns the directory used for caching Shinylive assets. This directory can contain
    multiple versions of Shinylive assets.
    """
    import appdirs

    return appdirs.user_cache_dir("shinylive")


def shinylive_assets_dir(version: str = SHINYLIVE_ASSETS_VERSION) -> str:
    """
    Returns the directory containing cached Shinylive assets, for a particular version
    of Shinylive.
    """
    return os.path.join(shinylive_cache_dir(), "shinylive-" + version)


def repodata_json_file(version: str = SHINYLIVE_ASSETS_VERSION) -> Path:
    return (
        Path(shinylive_assets_dir(version)) / "shinylive" / "pyodide" / "repodata.json"
    )


def copy_shinylive_local(
    source_dir: Union[str, Path],
    destdir: Optional[Union[str, Path]] = None,
    version: str = SHINYLIVE_ASSETS_VERSION,
):
    if destdir is None:
        destdir = Path(shinylive_cache_dir())

    destdir = Path(destdir)

    target_dir = destdir / f"shinylive-{version}"

    if target_dir.exists():
        shutil.rmtree(target_dir)

    shutil.copytree(source_dir, target_dir)


def ensure_shinylive_assets(
    destdir: Union[Path, None] = None,
    version: str = SHINYLIVE_ASSETS_VERSION,
    url: Optional[str] = None,
) -> Path:
    """Ensure that there is a local copy of shinylive."""

    if destdir is None:
        destdir = Path(shinylive_cache_dir())

    if url is None:
        url = shinylive_bundle_url(version)

    if not destdir.exists():
        print("Creating directory " + str(destdir))
        destdir.mkdir(parents=True)

    shinylive_bundle_dir = Path(shinylive_assets_dir(version))
    if not shinylive_bundle_dir.exists():
        print(f"{shinylive_bundle_dir} does not exist.")
        download_shinylive(url=url, version=version, destdir=destdir)

    return shinylive_bundle_dir


def remove_shinylive_assets(
    shinylive_dir: Union[str, Path, None] = None,
    version: Union[str, List[str], None] = None,
) -> None:
    """Removes local copy of shinylive.

    Parameters
    ----------
    shinylive_dir
        The directory where shinylive is stored. If None, the default directory will
        be used.

    version
        If a version is specified, only that version will be removed.
        If None, all local versions except the version specified by SHINYLIVE_ASSETS_VERSION will be removed.
    """

    if shinylive_dir is None:
        shinylive_dir = shinylive_assets_dir()

    shinylive_dir = Path(shinylive_dir)

    target_dir = shinylive_dir

    if isinstance(version, str):
        version = [version]

    if version is None:
        version = _installed_shinylive_versions(shinylive_dir)
        version = [re.sub("^shinylive-", "", os.path.basename(v)) for v in version]
        if SHINYLIVE_ASSETS_VERSION in version:
            print("Keeping version " + SHINYLIVE_ASSETS_VERSION)
            version.remove(SHINYLIVE_ASSETS_VERSION)

    target_dirs = [shinylive_dir / f"shinylive-{v}" for v in version]

    if len(target_dirs) == 0:
        print(f"No versions of shinylive to remove from {shinylive_dir}/")
        return

    for target_dir in target_dirs:
        print("Removing " + str(target_dir))
        if target_dir.exists():
            shutil.rmtree(target_dir)
        else:
            print(f"{target_dir} does not exist.")


def _installed_shinylive_versions(shinylive_dir: Optional[Path] = None) -> List[str]:
    if shinylive_dir is None:
        shinylive_dir = Path(shinylive_cache_dir())

    shinylive_dir = Path(shinylive_dir)
    subdirs = shinylive_dir.iterdir()
    subdirs = [re.sub("^shinylive-", "", str(s)) for s in subdirs]
    return subdirs


def print_shinylive_local_info() -> None:
    print(
        f"""    Local cached shinylive asset dir:
    {shinylive_cache_dir()}
    """
    )
    if Path(shinylive_cache_dir()).exists():
        print("""    Installed versions:""")
        installed_versions = _installed_shinylive_versions()
        if len(installed_versions) > 0:
            print("    " + "\n    ".join(installed_versions))
        else:
            print("    (None)")
    else:
        print("    (Cache dir does not exist)")


def _check_assets_url(
    version: str = SHINYLIVE_ASSETS_VERSION, url: Optional[str] = None
) -> bool:
    """Checks if the URL for the Shinylive assets bundle is valid.

    Returns True if the URL is valid (with a 200 status code), False otherwise.

    The reason it has both the `version` and `url` parameters is so that it behaves the
    same as `download_shinylive()` and `ensure_shinylive_assets()`.
    """
    if url is None:
        url = shinylive_bundle_url(version)

    req = urllib.request.Request(url, method="HEAD")
    resp = urllib.request.urlopen(req)
    status = resp.getcode()

    if status == 200:
        return True
    else:
        return False



import json
from pathlib import Path
from typing import Optional, Union

import click

from . import _assets, _deps, _export, _version


@click.group(
    no_args_is_help=True,
    help=f"""
    \b
    shinylive Python package version: {_version.SHINYLIVE_PACKAGE_VERSION}
    shinylive web assets version:     {_assets.SHINYLIVE_ASSETS_VERSION}
""",
)
def main() -> None:
    pass


@main.command(
    help="""
Turn a Shiny app into a bundle that can be deployed to a static web host.

APPDIR is the directory containing the Shiny application.

DESTDIR is the destination directory where the output files will be written to. This
directory can be deployed as a static web site.

This command will not include the contents of venv/ or any files that start with '.'

After writing the output files, you can serve them locally with the following command:

    python3 -m http.server --directory DESTDIR 8008
""",
    no_args_is_help=True,
)
@click.argument("appdir", type=str)
@click.argument("destdir", type=str)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="Print debugging information when copying files.",
    show_default=True,
)
@click.option(
    "--subdir",
    type=str,
    default=None,
    help="Subdir in which to put the app.",
    show_default=True,
)
@click.option(
    "--full-shinylive",
    is_flag=True,
    default=False,
    help="Include the full Shinylive assets, including all Pyodide packages. Without this flag, only the packages needed to run the application are included.",
    show_default=True,
)
def export(
    appdir: str,
    destdir: str,
    subdir: Union[str, None],
    verbose: bool,
    full_shinylive: bool,
) -> None:
    _export.export(
        appdir,
        destdir,
        subdir=subdir,
        verbose=verbose,
        full_shinylive=full_shinylive,
    )


@main.command(
    help=f"""Manage local copy of Shinylive web assets

    \b
    Commands:
      download: Download assets from the remote server.
      remove: Remove local copies of assets. By default, removes all versions except {_assets.SHINYLIVE_ASSETS_VERSION}. Can be used with --version to remove a specific version.
      info: Print information about the local assets.
      install-from-local: Install shinylive assets from a local directory. Must be used with --source.

""",
    no_args_is_help=True,
)
@click.argument("command", type=str)
@click.option(
    "--version",
    type=str,
    default=None,
    help="Shinylive version to download or remove.",
    show_default=True,
)
@click.option(
    "--url",
    type=str,
    default=None,
    help="URL to download from. If used, this will override --version.",
    show_default=True,
)
@click.option(
    "--dir",
    type=str,
    default=None,
    help="Directory to store shinylive assets (if not using the default)",
)
@click.option(
    "--source",
    type=str,
    default=None,
    help="Directory where shinylive assets will be copied from. Must be used with 'copy' command.",
)
def assets(
    command: str,
    version: Optional[str],
    url: Optional[str],
    dir: Union[str, Path],
    source: Optional[str],
) -> None:
    if dir is None:
        dir = _assets.shinylive_cache_dir()
    dir = Path(dir)

    if command == "download":
        if version is None:
            version = _version.SHINYLIVE_ASSETS_VERSION
        _assets.download_shinylive(destdir=dir, version=version, url=url)
    elif command == "remove":
        _assets.remove_shinylive_assets(shinylive_dir=dir, version=version)
    elif command == "info":
        _assets.print_shinylive_local_info()
    elif command == "install-from-local":
        if source is None:
            raise click.UsageError("Must specify --source")
        if version is None:
            version = _version.SHINYLIVE_ASSETS_VERSION
        print(f"Copying shinylive-{version} from {source} to {dir}")
        _assets.copy_shinylive_local(source_dir=source, destdir=dir, version=version)
    else:
        raise click.UsageError(f"Unknown command: {command}")


@main.command(
    help="""Get a set of base dependencies for a Shinylive application.

    This is intended for use by the Shinylive Quarto extension.
""",
    no_args_is_help=True,
)
@click.option(
    "--path-prefix",
    type=str,
    default="shinylive-dist/",
    help="A prefix to prepend to the `path` for each dependency.",
    show_default=True,
)
def base_deps(path_prefix: str) -> None:
    deps = _deps.shinylive_base_deps_htmldep(path_prefix)
    print(json.dumps(deps, indent=2))


@main.command(
    help="""Get a set of dependencies for a set of Python files packaged into a .json file.

    This is intended for use by the Shinylive Quarto extension.
""",
    no_args_is_help=True,
)
@click.argument("json_file", type=str)
def package_deps(json_file: str) -> None:
    deps = _deps.package_deps_htmldep(json_file)
    print(json.dumps(deps, indent=2))
